/*
    reader.c - OSS trampoline sample client
	
    Copyright (C) 2000 Ram Sivrasubramanium <ram@linux-workshop.com>
    
    Modifyed 2003 by Christian Wolff <scarab-dsp@scarabaeus.org>
    2003-08-01: added running time display on ioctl print
	
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
	
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
	
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include "dsproxy.h"

static int read_exact(
	int fd, 
	void *buf2, 
	size_t count) 
{
	char *buf = buf2;
	int savedcount = count;
	size_t siz;
	
	while (count > 0) {
		siz = read(fd, buf, count);
		if (siz < 0) return(siz);
		
		buf += siz;
		count -= siz;
	}
	
	return savedcount;
}


static int connect_to_dsproxy_server(
	char *hostname)
{
	struct hostent *hentry;
	struct sockaddr_in serv_add;
	unsigned int addr;
	int sock;
	
	hentry = gethostbyname(hostname);
	addr = *(unsigned int *)hentry->h_addr;
	
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		printf("Could not get socket!\n");
		exit(255);
	}
	
	serv_add.sin_family = AF_INET;
	serv_add.sin_port   = htons(DSPROXY_TCP_PORT);
	serv_add.sin_addr.s_addr = addr;
	bzero(&(serv_add.sin_zero), 8);
	
	if (connect(sock, (struct sockaddr *)&serv_add, sizeof(serv_add)) == -1) {
		printf("Could not connect to %s.\n",hostname);
		exit(255);
	}
	
	return(sock);
}

static void print_time(
	long long bytes, 
	int bytes_per_sec, 
	long long csec_offs) 
{
	long long h, m, s, f;
	f = (bytes * 100LL + csec_offs) / (long long)bytes_per_sec;
	s = f / 100LL;
	f = f % 100LL;
	m = s / 60LL;
	s = s % 60LL;
	h = m / 60LL;
	m = m % 60LL;
	fprintf(stderr, "%02llu:%02llu:%02llu.%02llu ", h, m, s, f);
}


int main(
	int argc, 
	char **argv)
{
	int fdi = -1, fdo = -1, fdo2 = -1, fdmix = -1;
	char buffer[DSPROXY_FIFO_SIZE + 4];  /* bytes is the size of a stereo 16-bit sample for coalesce purposes */
	char op;
	unsigned short len;
	int len2, cmd, val, i;
	
	int dump_ioctls = 0;
	int dump_english_ioctls = 0;
	int do_reset = 0;
	int do_stdout = 0;
	int do_file = 0;
	int disable_fragment_datapacing = 0;
	char *do_file_name = NULL;
	int x_dspx_output = 0;
	int offset = 0;
	char *hostname = NULL;
	char *sdevname;
	long long total_bytes = 0;
	long long centisecond_offs = 0;
	int bytes_per_sample = 2;
	int sample_per_sec = 44100;
	int channels = 2;
	int bytes_per_sec = 176400;
	
	for (i = 1; i < argc; i++) {
		if ((!strcmp(argv[i], "-h")) || (! strcmp(argv[i], "-help"))) {
			fprintf(stderr, "Usage:\n------\n");
			fprintf(stderr, "%s [options]\n\n", argv[0]);
			fprintf(stderr, 
				"-d\t\tdisable fragment datapacing\n"
				"-e\t\tdump \"english language\" ioctls to stderr\n"
				"-f filename\toutput pcm audio data to specified file\n"
				"-h\t\thelp\n"
				"-i\t\tdump ioctls to stderr\n"
				"-n hostname\tconnect to dsproxy server at hostname port %d\n"
				"-r\t\topen/close real dsp device when reset ioctl is encountered\n"
				"-s\t\toutput pcm audio data to stdout\n"
				"-x\t\tdon't output to /dev/dsp[_x]\n",
				
				DSPROXY_TCP_PORT
			);
			exit(0);
		} else if (!strcmp(argv[i], "-d")) {
			disable_fragment_datapacing = 1;
		} else if (!strcmp(argv[i], "-e")) {
			dump_english_ioctls = 1;
		} else if (!strcmp(argv[i], "-i")) {
			dump_ioctls = 1;
		} else if (!strcmp(argv[i], "-r")) {
			do_reset = 1;
		} else if (!strcmp(argv[i], "-s")) {
			do_stdout = 1;
		} else if (!strcmp(argv[i], "-f")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing filename after -f option!\n");
				exit(255);
			}
			do_file_name = argv[i];
			do_file = 1;
		} else if (!strcmp(argv[i], "-n")) {
			i++;
			if (i >= argc) {
				fprintf(stderr, "Missing filename after -n option!\n");
				exit(255);
			}
			hostname = argv[i];
		} else if (!strcmp(argv[i], "-x")) {
			x_dspx_output = 1;
		}
	}
	
	if (hostname) {
		fdi = connect_to_dsproxy_server(hostname);
	} else {
		fdi = open("/dev/dsp", O_RDONLY);
	}
	
	if (fdi < 0) {
		fprintf(stderr, "Error in opening %s!\n", (hostname) ? hostname : "/dev/dsp");
		exit(255);
	}
	
	if (! x_dspx_output) {
		if ((fdo = open(sdevname = hostname ? "/dev/dsp" : "/dev/dsp_x", O_WRONLY)) < 0) {
			fprintf(stderr, "Error in opening %s!\n", sdevname);
			exit(255);
		}
		
		if ((fdmix = open(sdevname = hostname ? "/dev/mixer" : "/dev/mixer_x", O_WRONLY)) < 0) {
			fprintf(stderr, "Error in opening %s!\n", sdevname);
			exit(255);
		}
	}
	
	if (do_file) {
		if ((fdo2 = creat(do_file_name, S_IRWXU)) < 0) {
			fprintf(stderr, "Error in opening `%s'!\n", do_file_name);
			exit(255);
		}
	}
	
	
	for (;;) {
		int prefixedlen;
		
		read_exact(fdi, &op, sizeof(char));
		switch (op) {
		case DSPROXY_STREAM_DSPDATA:	
			read_exact(fdi, &len, sizeof(short));
			len2 = read_exact(fdi, buffer + offset, (len == 0) ? 65536 : len);
			total_bytes += len2;
			
			if ((prefixedlen = len2 + offset) & 3) {  /* look if we have an odd # of bytes and emit only a multiple of 4 to the soundcard in order to keep from turning the /dev/dsp */
				int fixedlen = prefixedlen & (~3);  /* trampoline into an atari teenage riot transform function */
				if (! x_dspx_output) write(fdo, buffer, fixedlen);
				if (do_file) write(fdo2, buffer, fixedlen);
				if (do_stdout) write(STDOUT_FILENO, buffer, fixedlen);
				memmove(buffer, buffer + fixedlen, offset = prefixedlen & 3);  /* make "odd" (<4 byte) tail the new head to keep dsp aligned */
			} else {
				if (! x_dspx_output) write(fdo, buffer, prefixedlen);
				if (do_file) write(fdo2, buffer, prefixedlen);
				if (do_stdout) write(STDOUT_FILENO, buffer, prefixedlen);
				offset = 0;
			}
			if (dump_english_ioctls) {
				print_time(total_bytes, bytes_per_sec, centisecond_offs);
				fprintf(stderr, "\r");
			}
			break;
		
		case DSPROXY_STREAM_IOCTL:
			read_exact(fdi, &cmd, sizeof(int));
			read_exact(fdi, &val, sizeof(int));
			
			if (! x_dspx_output) {
				if ((cmd == SNDCTL_DSP_SETFRAGMENT) && (! disable_fragment_datapacing)) {
					int fragcnt = (val >> 16);
					int fragsiz = (val & 0xffff);
					
					if (fragcnt > 2) fragcnt = 2;
					val = (fragcnt << 16) | fragsiz;
				}
				ioctl(fdo, cmd, &val);
				
				if ((cmd == SNDCTL_DSP_RESET) && (! do_stdout) && (! do_file) && (do_reset)) {
					close(fdo);
					fdo = open("/dev/dsp_x", O_WRONLY);
				}
			}
			
			switch (cmd) {
			case SNDCTL_DSP_RESET:
				offset = 0;  /* what the user intends if he upswitches from U8 mono to X16 stereo to keep misalignments from happening */
				break;
			case SNDCTL_DSP_SPEED:
				if (sample_per_sec != val) {
					centisecond_offs += (total_bytes * 100LL / bytes_per_sec);
					total_bytes = 0;
				}
				sample_per_sec = val;
				bytes_per_sec = bytes_per_sample * sample_per_sec * channels;
				break;
			case SNDCTL_DSP_CHANNELS:
				if (channels != val) {
					centisecond_offs += (total_bytes * 100LL / bytes_per_sec);
					total_bytes = 0;
				}
				channels = val;
				bytes_per_sec = bytes_per_sample * sample_per_sec * channels;
				break;
			case SNDCTL_DSP_STEREO:
				if ((channels == 2) != (val != 0)) {
					centisecond_offs += (total_bytes * 100LL / bytes_per_sec);
					total_bytes = 0;
				}
				channels = val ? 2 : 1;
				bytes_per_sec = bytes_per_sample * sample_per_sec * channels;
				break;
			case SNDCTL_DSP_SETFMT:
				centisecond_offs += (total_bytes * 100LL/ bytes_per_sec);
				total_bytes = 0;
				if ((val & AFMT_U8) || (val & AFMT_S8)) bytes_per_sample = 1;
				else bytes_per_sample = 2;
				bytes_per_sec = bytes_per_sample * sample_per_sec * channels;
				break;
			}
			
			if (dump_ioctls) {
				fprintf(stderr, "ioctl: %08x %08x\n", cmd, val);
			}
			
			if (dump_english_ioctls) {
				print_time(total_bytes, bytes_per_sec, centisecond_offs);
				switch (cmd) {
				case SNDCTL_DSP_RESET:
					fprintf(stderr, "SNDCTL_DSP_RESET\n");
					break;
				
				case SNDCTL_DSP_POST:
					fprintf(stderr, "SNDCTL_DSP_POST\n");
					break;
				
				case SNDCTL_DSP_SPEED:
					fprintf(stderr, "SNDCTL_DSP_SPEED    => %d\n", val);
					break;
				
				case SNDCTL_DSP_STEREO:
					fprintf(stderr, "SNDCTL_DSP_STEREO   => %s\n", val ? "yes" : "no");
					break;
				
				case SNDCTL_DSP_CHANNELS:
					fprintf(stderr, "SNDCTL_DSP_CHANNELS => %d\n", val);
					break;
				
				case SNDCTL_DSP_SETFRAGMENT:
					fprintf(stderr, "SNDCTL_DSP_SETFRAGMENT => %08x\n", val);
					break;
				
				case SNDCTL_DSP_SETFMT:
					fprintf(stderr, "SNDCTL_DSP_SETFMT   =>");
					if (val & AFMT_U8) fprintf(stderr, " AFMT_U8");
					if (val & AFMT_S8) fprintf(stderr, " AFMT_S8");
					if (val & AFMT_S16_LE) fprintf(stderr, " AFMT_S16_LE");
					if (val & AFMT_S16_BE) fprintf(stderr, " AFMT_S16_BE");
					if (val & AFMT_S16_NE) fprintf(stderr, " AFMT_S16_NE");
					if (val & AFMT_U16_LE) fprintf(stderr, " AFMT_U16_LE");
					if (val & AFMT_U16_BE) fprintf(stderr, " AFMT_U16_BE");
					fprintf(stderr, "\n");
					break;
				}
			}
			break;
		
		case DSPROXY_MIXER_IOCTL:
			read_exact(fdi, &cmd, sizeof(int));
			read_exact(fdi, &val, sizeof(int));
			
			if (! x_dspx_output) 
				ioctl(fdmix, cmd, &val);
			
			if (dump_english_ioctls) {
				print_time(total_bytes, bytes_per_sec, centisecond_offs);
				switch (cmd) {
				case SOUND_MIXER_WRITE_VOLUME:
					fprintf(stderr, "SOUND_MIXER_WRITE_VOLUME => %d - %d\n", (val >> 8) & 0xff, val & 0xff);
					break;
				
				case SOUND_MIXER_WRITE_BASS:
					fprintf(stderr, "SOUND_MIXER_WRITE_BASS   => %d - %d\n", (val >> 8) & 0xff, val & 0xff);
					break;
				
				case SOUND_MIXER_WRITE_TREBLE:
					fprintf(stderr, "SOUND_MIXER_WRITE_TREBLE => %d - %d\n", (val >> 8) & 0xff, val & 0xff);
					break;
				
				case SOUND_MIXER_WRITE_PCM:
					fprintf(stderr, "SOUND_MIXER_WRITE_PCM    => %d - %d\n", (val >> 8) & 0xff, val & 0xff);
					break;
				
				case SOUND_MIXER_WRITE_OGAIN:
					fprintf(stderr, "SOUND_MIXER_WRITE_PCM    => %d - %d\n", (val >> 8) & 0xff, val & 0xff);
					break;
				}
			}
			
			break;
		}
	}
}

